clear
clc
close all
tic
warning off

if(ispc)
    nCores = str2double(getenv('NUMBER_OF_PROCESSORS'));
elseif(ismac)
    [status, output] = system('sysctl hw.ncpu | awk ''{print $2}''');
    nCores = str2double(output);
end
if(matlabpool('size') <= 1)
    matlabpool
end

global params
params.minFrac = 0.0;

colors = {'m.' 'bo' 'gs' 'cd' 'y^' 'kv' 'r*' 'w+'};
outlier_colors = {'mx','bx','gx','cx','yx','kx','rx','wx'};
n_inlier_structs = [];

epsilon = 0.1;
nIterScale = 4000;
nIter = 2000;
isGrassmann = 1;
isHeteroscedastic = 1;
isHomogeneous = 1;
kernelType = 'Gaussian'; % 'Epanechnikov'; % 
nRuns = 1;

% fname = 'Oxford_MertonCollege3_PtMatches_images';
fname = 'Oxford_MertonCollege1_PtMatches_SIFT_outliers';

load([fname,'.mat']);

debug = true;

x1 = pts1(:,1); y1 = pts1(:,2);
x2 = pts2(:,1); y2 = pts2(:,2);

nOutliers = 0;
x1_out = rand(nOutliers,1)*size(im1,2);
y1_out = rand(nOutliers,1)*size(im1,1);

x2_out = rand(nOutliers,1)*size(im1,2);
y2_out = rand(nOutliers,1)*size(im1,1);
x1 = [x1; x1_out]; y1 = [y1; y1_out]; x2 = [x2; x2_out]; y2 = [y2; y2_out];
gt(end+1:end+nOutliers) = 0;

%% ignore classes
classes2remove = [0 5 6]; % [ 0 1 2 3 4 5];
for c2r_i = classes2remove
    x1(gt == c2r_i) = [];
    y1(gt == c2r_i) = [];
    x2(gt == c2r_i) = [];
    y2(gt == c2r_i) = [];
    gt(gt == c2r_i) = [];
end
ground_truth = gt;
nStructs = sum(unique(gt) > 0);
labels = unique(gt(:))';


%%
X1 = [x1 y1 ones(size(x1,1),1)]';
X2 = [x2 y2 ones(size(x1,1),1)]';

[X1 T1] = normalise2dpts(X1);
[X2 T2] = normalise2dpts(X2);

Y = computeCarrierVectors_homography(X1,X2);

allData = Y;
allInliers = cell(nRuns,0);
allTheta = cell(nRuns,0);
allScale = cell(nRuns,0);
allAlpha = cell(nRuns,0);
allTransforms = cell(nRuns,0);
allLeftOverPts = cell(nRuns,1);
discr_results = cell(nRuns,2);


allC = computeCovarianceMatrices_homography(X1,X2);

fig_num = 324213;

for run_i = 1:nRuns,
    fprintf(1,'\n\n Run %d of %d\n\n',run_i,nRuns);
    n = size(Y,1);
    k = 1;
    nCarriers = 2;
   
    orig_inds = 1:length(x1);
    ground_truth = gt;
    best_norm_score = -inf;
    nElem = 4; % (n-k+1)/nCarriers;
  
    X = [X1; X2];
    num_min_inliers = params.minFrac*size(X,2);
    degen_func = @isdegenerate_homography;

    if(debug)
        figure(fig_num);subplot(2,2,4);imshow(im1);hold on;
        for ii = labels,
            if(ii == 0)
                plot(x1(gt == ii),y1(gt == ii),outlier_colors{1});
            else
                plot(x1(gt == ii),y1(gt == ii),colors{ii});
            end
        end
    end
    for ii = 1:10,
        params.minFrac = num_min_inliers/size(X,2);

%          C = calculateCovariances_homo(Y);
        
        tempX1 = X(1:3,:);
        tempX2 = X(4:6,:);
        
        [tempX1 T1] = normalise2dpts(tempX1);
        [tempX2 T2] = normalise2dpts(tempX2);
        allTransforms{run_i,ii} = [T1,T2];
        Y = computeCarrierVectors_homography(tempX1,tempX2);
        X = [tempX1;tempX2];
%         C = calculateCovariances_homo(Y);
        C = computeCovarianceMatrices_homography(tempX1,tempX2);
        
%% scale computation        
        fprintf('%d Generating hypotheses...',ii);

        parfor j = 1:nCores,
            hypsInScale_{j} = generateHypotheses(nIterScale/8, nElem, X,degen_func);
        end
        hypsInScale = [hypsInScale_{:}];

        fprintf('Done.\nComputing scale...');
        [scale frac bestHypS peakVal inliersScale bestTheta, enclosedPts enclosedFrac] = ...
            estimateScale(Y, n, k, hypsInScale, 0, nCarriers,C,nElem,isHomogeneous,epsilon,isHeteroscedastic);

        fprintf('Done.\n');
        fprintf('quantized frac = %f, est frac = %f, scale = %f, peakVal = %f, est inliers = %d, nInliersS = %d\n',...
            frac, enclosedFrac, scale, peakVal, enclosedPts, length(inliersScale));
        
        if(debug)
            % plot scale inliers and hypotheses
            figure(fig_num);subplot(2,2,1);
            if(ii == 1)
                imshow(im1);
            end
            hold on; plot(x1(orig_inds(bestHypS)),y1(orig_inds(bestHypS)),[colors{ii}(1),'o'],'markersize',8,'linewidth',3);
            title('scale inliers');
        
            outlier_indices = inliersScale(gt(orig_inds(inliersScale))==0);
            inlier_indices = inliersScale(gt(orig_inds(inliersScale))~=0);
            plot(x1(orig_inds(inlier_indices)),y1(orig_inds(inlier_indices)),colors{ii});
            plot(x1(orig_inds(outlier_indices)),y1(orig_inds(outlier_indices)),outlier_colors{ii});
        end
        allScale{run_i,ii} = scale;


%% pbM based model estimation
        fprintf('Generating hypotheses...');
        parfor j = 1:nCores,
            hypsIn_{j} = inliersScale(generateHypotheses(nIter/nCores, nElem, X(:,inliersScale),degen_func));
        end
        hypsIn = [hypsIn_{:}];
        
        hypsIn = hypsIn(:,1:size(hypsIn,2)-mod(size(hypsIn,2),nCores));
        hypsPerThread = size(hypsIn,2)/nCores;
        hyps = reshape(hypsIn,nElem,hypsPerThread,nCores);
        fprintf('Done.\nRunning pbM...');

        tic;

        parfor j = 1:nCores,
            [theta_(:,:,j) alpha_(:,j) score_(j) inliers_{j} bestHyp_(:,j)] = estimateModel_mex(Y, n, k, hyps(:,:,j),C, scale/2, kernelType, nCarriers,nElem,1,isGrassmann,isHeteroscedastic);
        end
        pbm_time = toc;
        
        fprintf('Done.\n');

        [score b] = max(score_);
        norm_score = score/(norm(scale/2))^2;
        if(best_norm_score < norm_score)
            best_norm_score = norm_score;
        end
        theta  = theta_(:,:,b);
        alpha = alpha_(:,b);
        score = score_(b);
        inliers = inliers_{b};
        bestHyp = bestHyp_(:,b);
        
        allTheta{run_i,ii} = theta;
        allAlpha{run_i,ii} = alpha;
        
        if(debug)
            figure(fig_num);subplot(2,2,2);
            if(ii == 1)
                imshow(im1);
            end
            hold on; plot(x1(orig_inds(bestHyp)),y1(orig_inds(bestHyp)),[colors{ii}(1),'o'],'markersize',6,'linewidth',3);
            title('pbM inliers');
            
            outlier_indices = inliers(gt(orig_inds(inliers))==0);
            inlier_indices = inliers(gt(orig_inds(inliers))~=0);
            plot(x1(orig_inds(inlier_indices)),y1(orig_inds(inlier_indices)),colors{ii});
            plot(x1(orig_inds(outlier_indices)),y1(orig_inds(outlier_indices)),outlier_colors{ii});
        end

        clear theta_ alpha_ score_ inliers_ bestHyp_


        fprintf('Done PbM. \nInlier structure # %d nInliers = %d. Score = %f\n', ii,length(inliers),score);
        fprintf('normalized score = %f\n', norm_score);
        
%% dichotomy using k-dimensional KDE        
        fprintf('\nComputing KDE... ');
     

        kde_inliers = dichotomize_mex(Y,theta,C,alpha,scale/2,nCarriers,kernelType,isHeteroscedastic);

        if(ii == 1)
            strengths = length(kde_inliers)*score/norm(scale);
        else
            if(0.05*max(strengths) > length(kde_inliers)*score/norm(scale))
                fprintf(1,'\n Discovered %d inlier structures. Stopping....\n',ii-1);
                break;
            end
            strengths = [strengths length(kde_inliers)*score/norm(scale)];
        end
        if(debug)
            % plotting inliers discovered from KDE
            figure(fig_num);subplot(2,2,3);
            if(ii==1)
                imshow(im1);
            end
            hold on;
            kde_outlier_indices = kde_inliers(gt(orig_inds(kde_inliers))==0);
            kde_inliers_indices = kde_inliers(gt(orig_inds(kde_inliers))~=0);
            plot(x1(orig_inds(kde_inliers_indices)),y1(orig_inds(kde_inliers_indices)),colors{ii});
            plot(x1(orig_inds(kde_outlier_indices)),y1(orig_inds(kde_outlier_indices)),outlier_colors{ii});
            plot(x1(orig_inds(bestHyp)),y1(orig_inds(bestHyp)),[colors{ii}(1),'o'],'markersize',6,'linewidth',3);
            title('kde dichotomy');
        end
        
        allInliers{run_i,ii} = orig_inds(kde_inliers);
        fprintf('Done dichotomy.\n nInliers = %d. Score = %f\n\n', length(kde_inliers),score);

        inds2del = nCarriers*(kde_inliers-1)+1;
        for i = 2:nCarriers
            inds2del = [inds2del;inds2del(1,:) + i-1];
        end
        inds2del = inds2del(:)';
        Y(:,inds2del) = [];
        X(:,kde_inliers) = [];
        orig_inds(kde_inliers) = [];
        ground_truth(kde_inliers) = [];
    end
    
    allLeftOverPts{run_i} = orig_inds;
    allInliers{run_i,5} = allLeftOverPts{run_i};
    errors(run_i,1) = computeErrors(gt,allInliers(run_i,:));
end
