close all; clear; clc; warning off

global ground_truth params
if(ispc)
    nCores = str2double(getenv('NUMBER_OF_PROCESSORS'));
elseif(ismac)
    [status, output] = system('sysctl hw.ncpu | awk ''{print $2}''');
    nCores = str2double(output);
end
if(matlabpool('size') <= 1)
    matlabpool
end
params.minFrac = 0.075;
params.is_cvpr = false;

kernelType = 'Epanechnikov';
p = dlmread('faceData_P00_128x128_D10.txt');
load('face_ground_truth.mat','ground_truth');


n = size(p,2); % dimension of the data
k = 18;

nIterScale = 5000;
nIter = 100;
        
pp = p;
nElem = n-k+1;

doneSegment = 0;
nclustersPrev = 0;
nCarriers = 1;
isHeteroscedastic = 0;
isHomogeneous = 0;
isGrassmann = 0;
degenFunc = @isdegenerate_dummy;
epsilon = 0.1;
debug = false;

nClusters = 100;
nRuns = 1;

totalError = zeros(nRuns,1);
for iter = 1:nRuns,
    num_min_inliers = size(p,1)*params.minFrac;
    clear pts_cluster
    nClustersPrev = nClusters;
    
    nclusters = 0;
    ii = 0;
    nInliers = inf;
    theta = [];
    facesDetected = [];
    facesFreq = [];
    while(1)
        ii = ii + 1;
        if(ii == 1)
            Y = p';
            ind = 1:size(p,1);
        else
            ind = 1:size(p,1);
            if(exist('pts_cluster'))
                for j = 1:length(pts_cluster),
                    ind = setdiff(ind,pts_cluster{j,:});
                end
            end
            Y = p(ind,:)';
            params.minFrac = num_min_inliers/size(Y,2);
        end
       
        if(size(Y,2) < nElem+1)
            break;
        end
        C = calculateCovariances_homo(Y);
       
        fprintf('[iter=%d cluster=%d] Estimating scale...',iter,ii);
        tic
        
        parfor j = 1:nCores,
            hypsInScale_{j} = generateHypotheses(ceil(nIterScale/nCores), nElem, Y, degenFunc);
        end
        hypsInScale = [hypsInScale_{:}];
        [scale frac bestHypS peakVal inliersScale bestTheta, enclosedPts enclosedFrac] = estimateScale(Y, n, k, hypsInScale, debug, nCarriers,C,nElem,isHomogeneous,epsilon,isHeteroscedastic);

        if(length(inliersScale) < 10)
            break;
        end
        toc1 = toc;

        fprintf('Done. Time spent = %0.2fs.\n', toc1);
        fprintf('[iter=%d cluster=%d] frac = %f, peakVal = %f, nInliers = %d\n',iter,ii,frac, peakVal, numel(inliersScale));
        tic
      
        fprintf('Generating hypotheses...');
        parfor j = 1:nCores,
            hypsInPBM_{j} = generateHypotheses(ceil(nIter/nCores), nElem, Y(:,inliersScale),degenFunc);
        end
        hypsInPBM = [hypsInPBM_{:}];
        
        hypsInPBM = hypsInPBM(:,1:size(hypsInPBM,2)-mod(size(hypsInPBM,2),nCores));
        hypsPerThread = size(hypsInPBM,2)/nCores;
        hyps = reshape(inliersScale(hypsInPBM),nElem,hypsPerThread,nCores);
        
        fprintf('[iter=%d motion=%d] Executing pbM...',iter,ii);
                
        parfor j = 1:nCores,
            [theta_(:,:,j) alpha_(:,j) score_(j) inliers_{j} bestHyp_(:,j)] = estimateModel_mex(Y, n, k, hyps(:,:,j),C, scale, kernelType, nCarriers,nElem,isHomogeneous,isGrassmann,isHeteroscedastic);
        end
       
        [score b] = max(score_);
        theta = theta_(:,:,b);
        alpha = alpha_(:,b);
        score = score_(b);
        inliersPBM = inliers_{b};
        toc1 = toc;
        fprintf('Done. Time spent = %0.2fs. nInliers = %d. Score = %f\n', toc1, length(inliersPBM),score);
        tic
        fprintf('[iter=%d cluster=%d] Computing KDE...',iter,ii);

        inliersKDE = dichotomize_mex(Y,theta,C,alpha,scale/2,nCarriers,kernelType,isHeteroscedastic);
        toc1 = toc;
        fprintf('Done. Time spent = %.2fs. nInliers = %d.\n', toc1,length(inliersKDE));

        inliers = inliersKDE;
        if(~isempty(ground_truth))
            disp([ind(inliers);ground_truth(ind(inliers))']);
            freq = hist(sort(ground_truth(ind(inliers))));
            unique_ind = unique(ground_truth(ind(inliers)));
            freq = freq(find(freq>0));
            [freq_max idx] = max(freq);
            freq(idx) = [];
            totalError(iter) = totalError(iter) + sum(freq);
            if(length(facesDetected(find(facesDetected == unique_ind(idx)))) == 0)
                facesDetected = [facesDetected unique_ind(idx)];
                facesFreq = [facesFreq freq_max];
            else
                idx_temp = find(facesDetected == unique_ind(idx));
                if(facesFreq(idx_temp) < freq_max)
                    totalError(iter) = totalError(iter) + facesFreq(idx_temp);
                    facesFreq(idx_temp) = freq_max;
                else
                    totalError(iter) = totalError(iter) + freq_max;
                end
            end
        end
        pts_cluster{ii,:} = ind(inliers);
        nclusters = nclusters + 1;
        nInliers = length(inliers);
    end
    
    nclusters = size(pts_cluster,1);
end
