
function result = gpbm_ProjMotionSeg(p,im)

if(nargin == 0)
    load('motionData.mat','p','im','ground_truth','params');
end

npts = size(p,1);

pp = p;
% isotropic scaling
p = MakeNonHomogenous(p);
p = IsotropicScaling(p,0);

%% setup parameters for gpbm
global params;


if(ispc)
    nCores = str2double(getenv('NUMBER_OF_PROCESSORS'));
elseif(ismac)
    [status, output] = system('sysctl hw.ncpu | awk ''{print $2}''');
    nCores = str2double(output);
end
if(matlabpool('size') <= 1)
    matlabpool
end

epsilon = params.epsilon;
nFrames = params.nFrames;
isGrassmann = params.isGrassmann;
% isHeteroscedastic = params.isHeteroscedastic;
isHomogeneous = params.isHomogeneous;
kernelType = params.kernelType;


nIterScale = params.nIterScale;
nIter = params.nIter;
nCarriers = params.nCarriers;
plotMotionsFlag = params.plotMotions;
motionDim = params.motionDim;
degenFunc = @isdegenerate_dummy;

minFracStart = params.minFracStart;
minInliersNum = ceil(npts*minFracStart);
minInliers = npts*minFracStart;
best_norm_score = -inf;

nElem = params.nElem; % size of elemental subset


%% estimation
for iter = 1:2 % first affine then projective
    if(iter == 1)
        isHeteroscedastic = 0;
    else
        isHeteroscedastic = 0;
    end
    structCnt = 0;
    params.minFrac = minFracStart;
    clear ptsMotion;
    while(true)
        structCnt = structCnt + 1;
        if(structCnt == 1)
            Y = p';
            ind = 1:size(p,1);
        else
            ind = 1:size(p,1);
            if(exist('ptsMotion','var'))
                for j = 1:size(ptsMotion,1),
                    ind = setdiff(ind,ptsMotion{j,:});
                end
            end
            Y = p(ind,:)';
            ppGlobal = pp(ind,:);
            params.minFrac = minInliersNum/size(Y,2);
        end
        n = size(p,2); % dimension of data
        k = n-motionDim;
        
        % isotropic scaling
        if(iter == 1)
            Y = IsotropicScaling(Y',false);
        else
            Y = IsotropicScaling(Y',true);
        end
        Y = Y';
        if(iter == 1 || ~isHeteroscedastic) % affine estimation
            C = calculateCovariances_homo(Y); % homoscedastic
        else % projective estimation
            C = calculateCovariances_motion_3D(Y); % heteroscedastic
        end
        
        % check if enough points are available
        if(size(Y,2) < max(minInliers,7))
            break;
        end
        tic;
        % scale estimation
        fprintf('Motion %d Generating hypotheses...',structCnt);
        parfor j = 1:nCores,
            hypsInScale_{j} = generateHypotheses(ceil(nIterScale/nCores), nElem, Y, degenFunc);
        end
        hypsInScale = [hypsInScale_{:}];
        fprintf('Done.\nComputing scale...');
        [scale frac bestHypS peakVal inliersScale bestTheta, enclosedPts enclosedFrac] = estimateScale(Y, n, k, hypsInScale, 0, nCarriers,C,nElem,isHomogeneous,epsilon,isHeteroscedastic);
        fprintf('Done.\n');
        fprintf('Quantized frac = %f, est frac = %f, peakVal = %f, est inliersPBM = %d, nInliersS = %d\n',frac, enclosedFrac, peakVal, enclosedPts, length(inliersScale));
        if(frac < 0)
            break;
        end
        scale(scale < 1e-8) = 1e-8;
        t_scale = toc;
        fprintf('\n Time taken in step 1 Scale estimation - %.2f seconds\n',t_scale);
        tic;
        % model estimation using pbM
        fprintf('Generating hypotheses...');
        parfor j = 1:nCores,
            hypsInPBM_{j} = generateHypotheses(ceil(nIter/nCores), nElem, Y(:,inliersScale),degenFunc);
        end
        hypsInPBM = [hypsInPBM_{:}];
        
        hypsInPBM = hypsInPBM(:,1:size(hypsInPBM,2)-mod(size(hypsInPBM,2),nCores));
        hypsPerThread = size(hypsInPBM,2)/nCores;
        hyps = reshape(inliersScale(hypsInPBM),nElem,hypsPerThread,nCores);
        
        fprintf('Done.\n[iter=%d motion=%d] Executing pbM...',iter,structCnt);
               
        parfor j = 1:nCores,
            [theta_(:,:,j) alpha_(:,j) score_(j) inliers_{j} bestHyp_(:,j)] = estimateModel_mex(Y, n, k, hyps(:,:,j),C, scale, kernelType, nCarriers,nElem,isHomogeneous,isGrassmann,isHeteroscedastic);
        end
        fprintf('Done.\n');
        t_pbm = toc;
        fprintf('\n Time taken in step 2 Model estimation - %.2f seconds\n',t_pbm);
        [score b] = max(score_);

        theta  = theta_(:,:,b);
        alpha = alpha_(:,b);
        inliersPBM = inliers_{b};
        
        clear theta_ alpha_ score_ inliers_ bestHyp_
        fprintf('Inlier structure # %d nInliers = %d. Score = %f\n', structCnt,length(inliersPBM),score);
        
        tic;
        % dichotomy using KDE
        fprintf('Computing KDE... ');
       
        inliersKDE = dichotomize_mex(Y,theta,C,alpha,scale,nCarriers,kernelType,isHeteroscedastic);
        skipFundMat{structCnt} = ind(find(vnorm(Y'*theta - repmat(alpha',size(Y,2),1),2)' < norm(scale)/8));
        t_kde = toc;
        fprintf('\n Time taken in step 3 Dichotomy - %.2f seconds\n',t_kde);
        
        % termination condition
        if(structCnt == 1)
            strengths = length(inliersKDE)*score/norm(scale);
        else
            if(0.005*min(strengths) > length(inliersKDE)*score/norm(scale))
                fprintf('All inlier structures found. Exiting...\n');
                break;
            end
            strengths = [strengths length(inliersKDE)*score/norm(scale)];
        end
        
        fprintf('Done.\n nInliers = %d. Score = %f\n\n', length(inliersKDE),score);
        
        Y(:,inliersKDE) = []; % delete detected inlier points
        
        ptsMotion{structCnt,1} = ind(inliersKDE);
    end
    
    
    if(iter == 1)
        ptsMotionAffine = ptsMotion;
        result.ptsMotion{1,1} = ptsMotionAffine;
        result.nMotionsAffine = structCnt;
    else
        ptsMotionProj = ptsMotion;
        result.ptsMotion{3,1} = ptsMotionProj;
        result.nMotionsProj = structCnt;
    end
    
    %%% Refine using Fundamental matrix computation and reprojection
    if(iter == 1 && size(ptsMotionAffine,1) > 1)
        result.ptsMotion{2,1} = refineUsingFundamentalMatrices(pp,ptsMotionAffine,nFrames,npts,skipFundMat);
    elseif(iter > 1 && size(ptsMotionProj,1) > 1)
        result.ptsMotion{4,1} = refineUsingFundamentalMatrices(pp,ptsMotionProj,nFrames,npts,skipFundMat);
    end
    %%% Projective estimation of depths
    if(iter == 1)
        p = updateDepthEstimates(pp,ptsMotion,nFrames);
    end
    %     nMotions = size(ptsMotion,1);
    if(plotMotionsFlag)
        if(iter == 1)
            ha = plotMotions(im,pp,result.ptsMotion{1,1});set(ha,'name','Affine');
            haf = plotMotions(im,pp,result.ptsMotion{2,1});if(~isnan(haf)),set(haf,'name','Affine + Fundamental');end
        else
            hp = plotMotions(im,pp,result.ptsMotion{3,1});set(hp,'name','Projective');
            hpf = plotMotions(im,pp,result.ptsMotion{4,1});if(~isnan(hpf)),set(hpf,'name','Projective + Fundamental');end
        end
    end
end
end