#ifndef _CONJUGATE_GRADIENT_H_
#define _CONJUGATE_GRADIENT_H_

#include <stdio.h>
#include <math.h>
#include <algorithm>
#include <mex.h>
#include "commons.h"
#include "matrixOperations.h"

#define min(a,b) ((a) <= (b) ? (a) : (b))
#define max(a,b) ((a) >= (b) ? (a) : (b))

#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
#define SHFT(a,b,c,d) (a) = (b); (b) = (c); (c) = (d);

static float maxarg1,maxarg2;
#define FMAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ? (maxarg1) : (maxarg2))

#define RGOLD 0.618034
#define CGOLD (1-RGOLD)
#define GLIMIT 100.0
#define MAXITER 100

#define GOLD 1.618034
#define TINY 1.0e-10

class conjugateGradient{

public:
	conjugateGradient();
	~conjugateGradient();
	void initializeData(double* data, int n_, int N_, int nCarriers_, int npts_, int k_, double* scale_, double* invH_best, kernelType kType_, bool isHeteroscedastic_,bool isHomogenous_);
	void clearData();
	void localOptimize(double* theta, int& nIter);
private:
	
	// calculates function and gradient.
	double gmComputeFuncAndGrad(double *Y, double *G);

	// function used in line minimization
	inline double gmComputeFuncAlongLine(double t);

	// routines for line minimization
	double	gmDirMin(double& ax, double& bx, double& cx);
	void	gmbrak(double& ax, double& bx, double& cx, double& fa, double& fb, double& fc);
	double	gmbrent(double ainit, double binit, double cinit, double tol, double& xmin);

	// conjugate gradient minimization
	void		doGrassmannConjugateGradient(double ftol, int& nIter);
	inline void geodesicMove(double *y, double *u, double *v, double *c, double *s, double *ty);
	inline void parallelTransport(double *y, double *u, double *v, double *s, double *cs, double *sn, double *tg, double *th);

	bool hasData;
	int n,k,p,npts, nCarriers;

	double *Y;
	double *scale;
	double *proj_best;
	double *proj_all;
	double *invH;
	double* invH_best;
	int N;

	//temporary stuff
	double* dists;
	int* ind_constraints;


	int theta_dim[2];
	int data_dim[2];
	int Y_dim[2];

	kernelType kType;

	// data structures for grassmann cg
	double *m_ThetaAlpha;		// (n+1)*k
	double *m_dbArrT;		// (n+1)*k
	double *m_dbArrW;		// (n+1)*k

	double *m_G;			// (n+1)*k
	double *m_dbArrH;		// (n+1)*k
	double *m_dbArrX;		// (n+1)*k

	double *m_U;			// n * k + k
	double *m_V;			// k * k
	double *m_S;			// k
	double *m_Cos;			// k
	double *m_Sin;			// k

	double *m_Vec1;			// k, (z-alpha)
	double *m_Vec2;			// k, invH * (z-alpha)
	double *m_Vec3;			// 1, (z-alpha)' * invH * (z-alpha)


	//miscellaneous
	// workspace for LAPACK functions and matrix multiplications
	double *m_dbArrWork;	// 5 * n
	double *m_dbArrWork0;	// n * k
	double *m_dbArrWork1;	// k * k
	int m_nLwork;			// 5 * n
	int m_nInfo;

	// some pointers...
	double *pTheta;
	double *pProj;

	bool isHeteroscedastic;
	bool isHomogeneous;
};

#endif
